/*
(1)
  Calculates autodiffusion and (if not FREEBC) structure factor from
    previously recorded data

  The structure factor is weighted by `atomic masses': edit the
    ble-file to use `scattering amplitudes' b_i instead!

  The autodiffusion comes in two versions:

    - negative option('f'): Reads configurations from files simname.###,
      where ### runs from |option('f')|
    + positive option('f'): reads SIMNAME.plb from frame option('f')

  in both cases, the last configuration/frame read is given by
    variable no (included); the calculations are performed for every
    noint's configuration, ALL configurations are, however, read
    because it is necessary to follow the atomic path in periodic
    boundary conditions

  dt must not change to report correctly times!

  LIKELY WRONG: Viscosity via Einstein-Helfand formula
    cook* should be run with option -n:
      cook* -n -y-1      to record SIMNAME.vlb in addition to SIMNAME.plb
      cook* -f -n [-g]   to analyze (cf. plb2diff)
  SIMNAME.v.cp : viscosity
    column k*3+l+1 : Lkl(0)-Lkl(t), where k,l={xyz}, and
      Lkl=SUM_i r_i,k v_i,l m_i / V
    column 10: 1/6 SUM_offdiag [Lkl(0)-Lkl(t)]^2
    viscosity eta=(slope of column 10)/dt.plb*V/T in prog.units
    (1 prog.unit = 1.3806503e-05 kg/m/s)
  More about viscosity:
    Reduced viscosity (e.g., Lennard-Jones) is
      eta* = eta*sigma^2/sqrt(m*eps)
    or
      eta = eta* * sqrt(M*eps)/sigma^2/NA

(2)
  Calculates cross-section
*/

#include "ground.h"
#define SDS
#include "alloc.h"
#include <time.h>
#include "simglob.h"
#include "ewald.h"
#include "simils.h"
#include "sfdx.h"
#include "cpmark.h"
#include "units.h"
#include "maxjump.h"
#include "statics.h"
#include "norm.h" /* int iscube(void) */
#include "asksig.h"

static double t0;

#ifdef CLUSTERS
#include "intermac.h"
#include "simdef.h"
#include "cluster.c" /* updated and fixed 9/2009 */
#endif

void diffSF(double dt,double dtcp,int no,int n0,int nspec) /********* diffSF */
{
  int nd=0,n,i,j,ns,doit,by=abs(option('f'));
  ToIntPtr a0=cfg[0],a1=cfg[1],a0raw,lasta0,lasta0raw,firsta1,lasta1,firsta0=cfg[GearOrder+1];
  vector dr,*firstr,*firstv,*r,*lastr,*rraw,*lastrraw,*v,*lastv,*rcenter;
  vector *rrow=rof(molec,a0->rp); /* r: access by No.s sites */
#ifndef FREEBC
  int kk=0,nsf=0;
  vector SFsumL={0,0,0}; /* mess: see also sumL */
  double sumw=0,sumwq=0;
#endif
  time_t starttime,stoptime;
  FILE *qcp,*mcp,*visccp=NULL;
#define NVCP 14
  float *rcp,vcp[NVCP];
  double **charge; /* copy of charges; note that si[].charge is rewritten! */
  double firstt=0;
  vector oldL,lastL; /* ? - probably the same */
  int ir=0;
  double sumV=0,sumL=0,sumL2=0,RvdW;
  int nL=0;
  struct maxjump_s
    maxjump= {{0,0,0},{-1,-1,-1},{-1,-1,-1},{-1,-1,-1}},
    max1jump={{0,0,0},{-1,-1,-1},{-1,-1,-1},{-1,-1,-1}};

  static vector viscj0[3],viscjl[3];

  VV(oldL,=box.L) /* just avoid warnings...*/

  measure=1; /* needed by Ewald/SF */

  if (n0<1) ERROR(("init=%d: initial frame expected (1=1st frame)",n0))
  if (no<n0) ERROR(("final frame no=%d <= inital frame init=%d",no,n0))

  if (option('t')) time(&starttime);

  if (option('f')<0) {
    loadcfg(n0,&RvdW); /* 1st pass */
    loadcfg(n0,NULL); /* 2nd pass */
    t0=t; }
  else {
    /* don't use readplayback(n0,..): readplayback(0,..) leaves file opened */
    if (n0<1) ERROR(("n0<1"))
    readplayback(-n0,1);

    if (dt==0) {
      prt("WARNING: dt.%s not specified, time unit for autodiffusion is frame stride",
          option('f')<0?"cfg":"plb");
      dt=1; }

    t0=n0*dt; }

  if (tau.P) {
    int i;

    loop (i,0,No.s) VV(rrow[i],/=box.L)
    sumL+=box.L[0];
    sumL2+=Sqr(box.L[0]);
    sumV+=box.L[0]*box.L[1]*box.L[2];
    nL++; }

  sdsalloc(lasta0,a0->size); sdscopy(lasta0,a0);
  sdscopy(firsta0,a0);
  VV(lastL,=box.L)
  if (tau.P) VO(lastL,=1)

  /* for viscosity only, but I'm too lazy... */
  sdsalloc(a0raw,a0->size);   sdscopy(a0raw,a0);
  sdsalloc(lasta0raw,a0->size); sdscopy(lasta0raw,a0);
  sdsalloc(lasta1,a1->size);  sdscopy(lasta1,a1);
  sdsalloc(firsta1,a0->size); sdscopy(firsta1,a1);

  alloczero(rcenter,No.s*sizeof(vector));

  {
    int n,i;

    alloczero(charge,sizeof(charge[0])*nspec);

    loop (n,0,nspec) {
      siteinfo_t *si=spec[n]->si;
      int ns=spec[n]->ns;

      alloc(charge[n],sizeof(*si)*ns);

      loop (i,0,ns) {
#ifdef POLAR
        if (si[i].chargepol) ERROR(("internal")) /* -a0 forced by -f */
#endif
        charge[n][i]=si[i].charge; } } }

#ifndef FREEBC
  put2(el.alpha,el.kappa)
  /* initialize Ewald for structure factor */

  if (el.kappa<=0) {
    if (el.sf) WARNING(("el.sf=%d ignored because el.kappa=%g",el.sf,el.kappa))
    el.sf=0; }

  if (el.sf>0) {
    int nn;
    /* ugly patch to force Ewald to use weights instead of charges */
    sumw=sumwq=0;
    loop (nn,0,No.N) {
      molecule_t *mn=molec+nn;
      siteinfo_t *si=spec[mn->sp]->si;
      int ns=mn->ns;

      loop (i,0,ns) {
        sumw+=si[i].charge=si[i].sfweight;
        sumwq+=Sqr(si[i].charge); } }
    prt("SF: sum weights=%g  sum weights^2=%g",sumw,sumwq);

    if (FROM) WARNING(("option -j and SF"))

    /* the structure factor - initialize */
    if (el.sf==1) {
      if (!iscube()) ERROR(("sphericalized structure factor can be calculated for a cube only (try el.sf=3)"))
      el.sfsize=kk=(int)(Sqr(el.kappa*box.L[2])+1.000001);
      allocarrayzero(el.sfr,kk); }
    else if (el.sf==3) {
      /* upper estimate - WASTING MEMORY! */
      kk=(int)(el.kappa*box.L[0]+1.000001)
        *(int)(el.kappa*box.L[1]+1.000001)
        *(int)(el.kappa*box.L[2]+1.000001);
      allocarrayzero(el.sf3d,el.sfsize=kk*4); }

    if (tau.P) {
      VV(oldL,=box.L)
      /* WARNING: this will fail probably if not cube ... */
      if (!iscube()) ERROR(("sorry, variable non-cubic box not supported"))
      VO(box.L,=1) VO(box.Lh,=0.5)
      Ewald(0,NULL,NULL);
      VV(box.L,=oldL) VV(box.Lh,=0.5*box.L) }
    else
      /* fixed box size */
      Ewald(0,NULL,NULL); }
#endif

  mcp=fopen(Fn("m.cp"),"wb");
  qcp=fopen(Fn("q.cp"),"wb");
  if (option('n')) visccp=fopen(Fn("v.cp"),"wb");

  allocarrayzero(rcp,nspec+1);
  rcp[0]=CPmark;
  *(int4*)(rcp+1)=nspec+1;
  loop (i,2,nspec) sprintf((char*)(rcp+i),"%d",i);
  if (nspec>1) copy(rcp+nspec,"mdif",4); /* WARNING: no header for nspec=1 */
  fwrite(rcp,sizeof(*rcp),nspec+1,mcp);
  if (nspec>1) copy(rcp+nspec,"cond",4); /* WARNING: no header for nspec=1 */
  fwrite(rcp,sizeof(*rcp),nspec+1,qcp);
  if (option('n')) {
    int k1,k2;

    memset(vcp,' ',sizeof(vcp));
    loop (k1,0,3) loop (k2,0,3)
      sprintf((char*)(vcp+k1*3+k2),"v%c%c",k1+'x',k2+'x');
    vcp[0]=CPmark;
    *(int4*)(vcp+1)=NVCP;
    copy(vcp+9,"v000vmidvis0vislvcon",(NVCP-9)*4);
    fwrite(vcp,sizeof(vcp),1,visccp); }

  header(tau.P ? " t-t0[ps] t[ps]       msd           mscd     L [A]  "
               : " t-t0[ps] t[ps]       msd [A^2]     mscd     L [A]  ");

  loopto (n,n0,no) {
    if (sig) break;
    doit=(n-n0)%by==0;

    if (n>n0) { /* n=n0 already read */
      if (option('f')<0) {
        loadcfg(n,&RvdW); /* 1st pass */
        loadcfg(n,NULL); /* 2nd pass */
        if (fabs((n-n0)*dt+t0-t)>1e-4)
          WARNING(("linear time=%g, time read=%g",(n-n0)*dt+t0,t)) }
      else {
        readplayback(0,1);
        t=n*dt; }

      if (tau.P) {
        int i;

        loop (i,0,No.s) VV(rrow[i],/=box.L)
        if (!iscube()) ERROR(("not cube - 2nd"))
        sumL+=box.L[0];
        sumL2+=Sqr(box.L[0]);
        sumV+=box.L[0]*box.L[1]*box.L[2];
        nL++; }

      fprintf(stderr,"\r%d:%d ",n,no); }

#ifndef FREEBC
    if (doit) if (el.sf) {
      if (tau.P) {
        /* WARNING: box scaled to 1-cube, structure factor = ??? */
        VV(oldL,=box.L)
        if (!iscube() && el.sf==1)
          ERROR(("variable non-cubic box not supported for el.sf=1"))
        VO(box.L,=1) VO(box.Lh,=0.5)

        /* the structure factor */
        Ewald(1,NULL,a0->rp);
        Ewald(2,NULL,a0->rp);

        VV(box.L,=oldL) VV(box.Lh,=0.5*box.L) }
      else {
        /* the structure factor - fixed box */
        Ewald(1,NULL,a0->rp);
        Ewald(2,NULL,a0->rp); }

      nsf++;
      VV(SFsumL,+=box.L); /* for SF: cube only */

      if (option('t')) {
        time(&stoptime);
        fprintf(stderr,"%i/%i SF done at %lu = %s",
                nsf,n,         stoptime,ctime(&stoptime)); } }
#endif

#ifdef CLUSTERS
    if (doit) analyzeclusters(a0,n);
#endif

    /* remove nearest image jumps in the molecule movements */
#ifndef FREEBC
    if (tau.P) { VV(oldL,=box.L) VO(box.L,=1) VO(box.Lh,=0.5) }

    sdscopy(a0raw,a0);

    loop (j,FROM,No.N) {
      firstr=rof(molec+j,firsta0->rp);
      r=     rof(molec+j,a0->rp);
      lastr= rof(molec+j,lasta0->rp);
      ns=molec[j].ns;

      loop (i,0,ns) {
        int k;
        double D;

        loop (k,0,3) {
          D=r[i][k]/box.L[k]-lastr[i][k]/lastL[k];
          while (D>0.5)  { r[i][k]-=box.L[k]; D-=1; }
          while (D<-0.5) { r[i][k]+=box.L[k]; D+=1; }
          /* now r (a0) moves continuously over periodic boxes */

          if (fabs(D)>fabs(max1jump.xi[k])) {
            max1jump.xi[k]=D;
            max1jump.frame[k]=n;
            max1jump.n[k]=j;
            max1jump.i[k]=i;
            max1jump.init=init; max1jump.no=no; max1jump.f=option('f'); }

          D=r[i][k]/box.L[k]-firstr[i][k]/lastL[k];
          if (fabs(D)>fabs(maxjump.xi[k])) {
            maxjump.xi[k]=D;
            maxjump.frame[k]=n;
            maxjump.n[k]=j;
            maxjump.i[k]=i;
            maxjump.init=init; maxjump.no=no; maxjump.f=option('f'); } } } }

      if (tau.P) {
        VV(box.L,=oldL)
        VV(box.Lh,=0.5*box.L) }
#endif

    if (doit) {
      struct msd_s {
        double m,q;
        int n; } *msd,*msp;
      vector msdm,msdq;
      vector visc[3];
      vector visccont[3];
      double totm,totq;
      siteinfo_t *si;
      int sp,k;

      allocarrayzero(msd,nspec);
      VVO(msdm,=msdq,=0)
      loop (k,0,3) VO(visc[k],=0)
      loop (k,0,3) VO(visccont[k],=0)

      loop (j,FROM,No.N) {
        vector vm,vq;
        double summ=0;
        int k1,k2;

        VVO(vm,=vq,=0)
        firstr=rof(molec+j,firsta0->rp);
        r=rof(molec+j,a0->rp);
        rraw=rof(molec+j,a0raw->rp);
        lastrraw=rof(molec+j,lasta0raw->rp);
        firstv=rof(molec+j,firsta1->rp);
        v=rof(molec+j,a1->rp);
        lastv=rof(molec+j,lasta1->rp);
        ns=molec[j].ns;
        sp=molec[j].sp;
        msp=msd+sp;
        msp->n++;
        si=spec[sp]->si;

        loop (i,0,ns) {
          VVV(dr,=r[i],-firstr[i])

          VV(vq,+=charge[sp][i]*dr)
          VV(vm,+=si[i].mass*dr)
          summ += si[i].mass;

          if (option('n')) {
            loop (k1,0,3) {
              double DL=0;
              if      (rraw[i][k1]-lastrraw[i][k1]> box.Lh[k1]) DL= box.L[k1];
              else if (rraw[i][k1]-lastrraw[i][k1]<-box.Lh[k1]) DL=-box.L[k1];

              DL=-DL;

              loop (k2,0,3) {
                visc[k1][k2]+=(rraw[i][k1]*v[i][k2]-firstr[i][k1]*firstv[i][k2])*si[i].mass;
                visccont[k1][k2]+=(r[i][k1]*v[i][k2]-firstr[i][k1]*firstv[i][k2])*si[i].mass;
                viscj0[k1][k2]+=DL*v[i][k2]*si[i].mass;
                viscjl[k1][k2]+=DL*lastv[i][k2]*si[i].mass; } } } }

        msp->m+=SQR(vm)/Sqr(summ)/6;
        msp->q+=SQR(vq)/6;
        VV(msdq,+=vq)
        VV(msdm,+=vm) } /* molecules */

      loop (i,0,No.s) VV(rcenter[i],+=rrow[i])
      nd++;

      totm=SQR(msdm)/6;
      totq=SQR(msdq)/6;

      prt("%7.3f %8.3f %12.5g %12.5g (%9.5f,%9.5f,%9.5f)",
          (n-1)*dt,ir*dtcp+firstt,totm,totq,VARG(box.L));
      loop (i,0,nspec) rcp[i]=msd[i].m/(msd[i].n+(msd[i].n==0));
      rcp[nspec]=totm;
      fwrite(rcp,sizeof(*rcp),nspec+1,mcp);

      loop (i,0,nspec) rcp[i]=msd[i].q;
      rcp[nspec]=totq;
      fwrite(rcp,sizeof(*rcp),nspec+1,qcp);
      free(msd);

      if (option('n')) {
        int k1,k2,iii;
        double V=box.L[0]*box.L[1]*box.L[2];

        loop (iii,9,NVCP) {
          loop (k1,0,3) loop (k2,0,3) {
            if (iii==10) visc[k1][k2]+=viscj0[k1][k2];
            if (iii==11) visc[k1][k2]+=-viscj0[k1][k2]+viscjl[k1][k2];
            if (iii==12) visc[k1][k2]+=(viscj0[k1][k2]-viscjl[k1][k2])/2;
            vcp[3*k1+k2]=visccont[k1][k2]/V; }
          vcp[iii]=(Sqr(visc[0][1])+Sqr(visc[1][0])
                   +Sqr(visc[1][2])+Sqr(visc[2][1])
                   +Sqr(visc[2][0])+Sqr(visc[0][2]))/(12*Sqr(V)); }

        vcp[NVCP-1]=(Sqr(visccont[0][1])+Sqr(visccont[1][0])
                    +Sqr(visccont[1][2])+Sqr(visccont[2][1])
                    +Sqr(visccont[2][0])+Sqr(visccont[0][2]))/(12*Sqr(V));

        fwrite(vcp,sizeof(vcp),1,visccp); } }

    sdscopy(lasta0,a0);
    sdscopy(lasta1,a1);
    sdscopy(lasta0raw,a0raw);
    VV(lastL,=box.L)
    if (tau.P) VO(lastL,=1) }

  header("");
  fclose(qcp);
  fclose(mcp);
  if (option('n')) fclose(visccp);

  _n

    //  if (maxjump.i[0]>=0 || maxjump.i[1]>=0 || maxjump.i[2]) {
  /* now ALWAYS prints SIMNAME.aux, even if there are zeros there */
  {
    FILE *aux=fopen(Fn("aux"),"ab");
    int k;

    loop (k,0,3) if (maxjump.i[k]>=0) {
      prt("%c-axis:\n\
max jump over periodic boxes = %g*L (frames 1->%d, mol.s=%d.%d)",
          k+'x',maxjump.xi[k],maxjump.frame[k],maxjump.n[k],maxjump.i[k]);
      prt("max jump between consecutive frames: %g*L (frames %d->%d, mol.s=%d.%d)",
          max1jump.xi[k],max1jump.frame[k]-1,max1jump.frame[k],max1jump.n[k],max1jump.i[k]);
      if (fabs(max1jump.xi[k])>0.4)
        WARNING(("\
max jump between consecutive frames is too close to L/2\n\
*** diffusions etc. may be wrong - check whether msd and column difm are tiny\n\
*** if wrong, use lower -f# or re-run with shorter dt.plb")) }

    fwrite(&maxjump,sizeof(maxjump),1,aux);
    fwrite(&max1jump,sizeof(max1jump),1,aux);
    fclose(aux); }

  if (tau.P) prt("multiplication factor for diffusions and conductivities: <L^2>=%f\n\
  <L>^2=%f <V>^2/3=%f",sumL2/nL,Sqr(sumL/nL),pow(sumV/nL,2./3));

  {
    FILE *center=fopen("center.plb","wb");
    float fr[3];
    float q;

    fr[0]=No.s;
    if (tau.P) {
      fr[1]=sqrt(sumL2/nL);
      q=fr[1]/nd; }
    else {
      fr[1]=-3;
      q=1./nd; }
    fwrite(fr,4,2,center);
    VV(fr,=box.L)
    fwrite(fr,4,3,center);
    loop (i,0,No.s) {
      VV(fr,=q*rcenter[i])
      fwrite(fr,4,3,center); }
    fclose(center);
  }

#ifndef FREEBC
  if (el.sf==1) {
    FILE *SF=fopen(Fn("sfr"),"wt");
    double W=0;

    VO(SFsumL,/=nsf)
    fprintf(SF,"# sphericalized (radial) structure factor\n");
    fprintf(SF,"# %d frames read, <L>=%.8f\n",nsf,SFsumL[2]);
    fprintf(SF,"# k=1/wavelength (NB: in older version, there was factor 2 pi)\n");
    fprintf(SF,"# k/A^-1     S(k)      smooth[3]   smooth[5]   smooth[7]\n");
    if (option('v')&4) header(" k^2    S(k)     ");
    loop (i,1,kk)
      if (el.sfr[i].nk) Max(W,el.sfr[i].q/el.sfr[i].nk)
    loop (i,1,kk)
      if (el.sfr[i].nk) {
 /*.....      double sfi=el.sfr[i].q/(el.sfr[i].nk*sumwq);*/
        double sfi=1+No.s*(el.sfr[i].q/el.sfr[i].nk-sumwq)/Sqr(sumw);
        static double W1[3]={0.25,0.5,0.25};
        static double W2[5]={0.0625,0.25,0.375,0.25,0.0625};
        static double W3[7]={0.015625,0.09375,0.234375,0.3125,0.234375,0.09375,0.015625};
        double sm1,sm2,sm3;
        double qq,nnk;
        int is;

#define DOIT(M) \
        qq=nnk=0; \
        loopto (is,-M,M) {\
          int j=i+is; \
          if (j>=1 && j<kk) qq+=el.sfr[j].q*W##M[is+M],nnk+=el.sfr[j].nk*W##M[is+M]; }\
          sm##M=1+No.s*(qq/nnk-sumwq)/Sqr(sumw);

        DOIT(1)
        DOIT(2)
        DOIT(3)

        /* SF: cube only ... */
        fprintf(SF,"%8.5f %11.7f %11.7f %11.7f %11.7f %d\n",
                sqrt((double)i)/SFsumL[2],sfi,sm1,sm2,sm3,el.sfr[i].nk);
        if (option('v')&4) {
          prt_("%3i %9.6f ",i,sfi);
          graph(el.sfr[i].q/el.sfr[i].nk/W,65); } }
      else
        if (option('v')&4) prt("%3i",i);
    header("");
    fclose(SF); }

  if (el.sf==3) {
    FILE *SF=fopen(Fn("sf3d"),"wt");

    VO(SFsumL,/=nsf)
    fprintf(SF,"# 3D structure factor\n");
    fprintf(SF,"# %d frames read, <L>=%.8f %.8f %.8f\n",nsf,VARG(SFsumL));
    fprintf(SF,"# k=(unit vector of direction)/wavelength\n");
    fprintf(SF,"#kx/A^-1  ky/A^-1  kz/A^-1   k/A^-1       S(k)      nk\n");
    loop (i,0,kk*4)
      if (SQR(el.sf3d[i].k)) {
        double sfi=1+No.s*(el.sf3d[i].q/nsf-sumwq)/Sqr(sumw);

        fprintf(SF,"%8.5f %8.5f %8.5f  %8.5f  %11.7f  %d\n",
                el.sf3d[i].k[0]/SFsumL[0],
                el.sf3d[i].k[1]/SFsumL[1],
                el.sf3d[i].k[2]/SFsumL[2],
                sqrt(Sqr(el.sf3d[i].k[0]/SFsumL[0])+Sqr(el.sf3d[i].k[1]/SFsumL[1])+Sqr(el.sf3d[i].k[2]/SFsumL[2])),
                sfi,nsf); }
    header("");
    fclose(SF); }
#endif

#ifdef CLUSTERS
  bonddynamics();

  if (clhead) {
    cluster_t *c;
    FILE *removescript=fopen(Fn("rm.sh"),"wt");

    header(" total  formula   charge [name] bonds (0 = 1st molecule from chem.formula..)");
    for (c=clhead; c; c=c->next) prtcluster(removescript,c);
    if (clustcp) fclose(clustcp);
    fclose(removescript);
    header(""); }
#endif

  StaPrintAll(NULL);

  if (option('t')) {
    time(&stoptime);
    fprintf(stderr,"duration = %ld s = %.2f m\n",
            stoptime-starttime,(stoptime-starttime)/60.0); }
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#ifdef XSECTION

#include "simdef.h"
#include "rndgen.h"

#include "sphint.c"

typedef vector matrix[DIM];

static void rndmat(matrix M) /*************************************** rndmat */
/* random orientational matrix, orthonormal */
{
  double rr;

  rndonsph(M[0]);
  do rr=rndsph(M[2]); while (fabs(SCAL(M[0],M[2]))<0.0625);
  VECT(M[1],M[2],M[0])
  rr=sqrt(SQR(M[1]));
  VO(M[1],/=rr);
  VECT(M[2],M[0],M[1])
}

static void mpl(vector a,matrix M,vector b) /*************************** mpl */
/* a:=M.b */
{
  int i;

  loop (i,0,DIM) a[i]=SCAL(M[i],b);
}

#define XSECTION_M 1
/* molecule-based version */
#include "xsection.c"

#undef XSECTION_M
#define XSECTION_M 0
/* whole configuration version */
#include "xsection.c"

#endif
